from typing import Text, Dict, Any
import hashlib
import json
from sqlalchemy.orm import Session
import pytest
from rasa.nlu.training_data.formats import MarkdownReader
from rasax.community import metrics
from rasax.community import config
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.model_service import ModelService
from rasax.community.services.event_service import EventService
from rasax.community.constants import (
    CONFIG_METRICS_ENABLED,
    CONFIG_METRICS_USER,
    COMMUNITY_PROJECT_NAME,
)
from rasax.community.version import __version__
from tests.unit.conftest import insert_test_model, _user_event
from tests.unit.services.test_domain_service import TEST_DOMAIN
from tests.unit.services.test_data_service import EXAMPLE_DATA_MD
from tests.unit.services.test_story_service import TEST_STORY

METRICS_TEST_USER = "metrics-test-user"


@pytest.fixture
def enable_metrics(monkeypatch):
    new_config = {CONFIG_METRICS_ENABLED: True, CONFIG_METRICS_USER: METRICS_TEST_USER}
    monkeypatch.setattr(config, "metrics_collection_config", new_config)


@pytest.fixture
def metrics_events(enable_metrics, monkeypatch):
    events = []

    def fn(_, event_name, properties=None, context=None):
        if properties:
            properties.pop("metrics_id", None)

        events.append(
            {"event": event_name, "properties": properties, "context": context}
        )

    monkeypatch.setattr(metrics, "_track_internal", fn)
    return events


def test_get_status_event_empty(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
):
    actual = metrics._get_status_event(session)

    assert actual == {
        "project": hashlib.sha256(COMMUNITY_PROJECT_NAME.encode("utf-8")).hexdigest(),
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 0,
        "num_templates": 0,
        "num_slots": 0,
        "num_forms": 0,
        "num_intents": 0,
        "num_entities": 0,
        "num_stories": 0,
        "num_conversations": 0,
        "num_models": 0,
        "num_lookup_table_files": 0,
        "num_lookup_table_entries": 0,
        "num_synonyms": 0,
        "num_regexes": 0,
    }


async def test_get_status_event(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
    nlg_service: NlgService,
    event_service: EventService,
    new_project: Text,
    user: Dict[Text, Any],
    monkeypatch,
):
    monkeypatch.setattr(config, "project_name", new_project)

    domain_service.store_domain(TEST_DOMAIN, new_project)
    domain_id = domain_service.get_domain_id(new_project)
    assert domain_id is not None

    nlg_service.replace_templates(
        [{"template": "utter_greet", "text": "Hi there!"}], user["username"], domain_id
    )

    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    model_service = ModelService("path", session)
    insert_test_model(new_project, "metrics_test_model", model_service)

    await story_service.save_stories(TEST_STORY, user["team"])

    test_event = _user_event("metrics_test_user")
    event_service.save_event(json.dumps(test_event))

    session.commit()

    actual = metrics._get_status_event(session)

    assert actual == {
        "project": hashlib.sha256(new_project.encode("utf-8")).hexdigest(),
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 1,
        "num_templates": 1,
        "num_slots": 3,
        "num_forms": 1,
        "num_intents": 4,
        "num_entities": 2,
        "num_stories": 2,
        "num_conversations": 1,
        "num_models": 1,
        "num_lookup_table_files": 1,
        "num_lookup_table_entries": 15,
        "num_synonyms": 2,
        "num_regexes": 1,
    }


@pytest.mark.parametrize(
    "uuid, enabled", [("1234", True), ("5432", False), ("12344", True)]
)
def test_set_metrics_config_from_env_with_uuid(uuid: Text, enabled: bool):
    config.metrics_consent_given = enabled
    config.metrics_user_id = uuid
    metrics.set_metrics_config_from_env()

    assert metrics.are_metrics_enabled() == enabled
    assert metrics.metrics_id() == uuid


def test_set_metrics_config_from_env_without_uuid():
    config.metrics_user_id = None
    metrics.set_metrics_config_from_env()

    assert len(metrics.metrics_id()) >= 32


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/talk", "talk_to_your_bot"),
        ("http://localhost/talk/strict", "interactive_learning"),
        ("http://localhost/talk/lenient", "talk_to_your_bot"),
        ("http://localhost/talk/lenient/1567512149.2726388", "talk_to_your_bot"),
        ("http://localhost/stories", "stories"),
    ],
)
def test_track_story_created(metrics_events, referrer, origin):
    metrics.track_story_created(referrer)

    assert metrics_events[0] == {
        "event": "Story Created",
        "properties": {"story_created_from": origin},
        "context": None,
    }


@pytest.mark.parametrize(
    "username, input_channel",
    [
        ("me", "rasa"),
        ("test-user", "rasa"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", "Tester"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", None),
    ],
)
def test_track_message_received(metrics_events, username, input_channel):
    metrics.track_message_received(username, input_channel)

    assert metrics_events[0] == {
        "event": "Message Received",
        "properties": {
            "username": hashlib.sha256(username.encode("utf-8")).hexdigest(),
            "channel": input_channel or "rasa",
        },
        "context": None,
    }


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/conversations", "conversations"),
        (
            "http://localhost/conversations/bcf7bb85-0506-42fb-89d4-5c1e310c4133/13.16",
            "conversations",
        ),
        ("http://localhost/data", "annotate_new_data"),
    ],
)
def test_track_message_annotated_from_referrer(metrics_events, referrer, origin):
    metrics.track_message_annotated_from_referrer(referrer)

    assert metrics_events[0] == {
        "event": "Message Annotated",
        "properties": {"message_annotated_from": origin},
        "context": None,
    }


def test_track_message_annotated(metrics_events):
    metrics.track_message_annotated(metrics.MESSAGE_ANNOTATED_CONVERSATIONS)

    assert metrics_events[0] == {
        "event": "Message Annotated",
        "properties": {
            "message_annotated_from": metrics.MESSAGE_ANNOTATED_CONVERSATIONS
        },
        "context": None,
    }


def test_track_metrics_disabled(metrics_events):
    config.metrics_collection_config[CONFIG_METRICS_ENABLED] = False
    metrics.track("Foobar")

    assert len(metrics_events) == 0


def test_track_metrics_without_user(metrics_events):
    config.metrics_collection_config[CONFIG_METRICS_USER] = None
    metrics.track("Foobar")

    assert len(metrics_events) == 0
